function ECHAIM, lat_temp,lon_temp,dates_use, alts_in, user_hmF2 = user_hmF2, user_NmF2 = user_nmF2, $
          STORM_nmF2 = STORM_nmF2, satellite = satellite, update = update, forceUpdate = forceUpdate, $
          silence = silence, precip=precip, dregion=dregion

; lat_temp: The geographic latitude of the point of interest (can be a vector)
; lon_temp: The geographic longitude of the point of interest (can be a vector)
; dates_use: The UTC Julian times of interest (can use JULDAY(MONTH,DAY,YEAR,HOUR,MINUTE,SECOND) in place of this variable) (can be a vector)
; user_nmF2 and user_hmF2 are optional inputs of hmF2 and NmF2, which will override that used by E-CHAIM. hmF2 is in km and NmF2 is in e/m^3.
;                         These should be vectors of the same length as lat,lon,and dates_use.
; STORM_nmF2: Keyword forces the use of the NMF2 storm perturbation model for the NmF2 output and profile generation. 
; /satellite: This keyword should be set if a user is interested in a time-varying set of altitudes and locations.
;             If this keyword is set, alts_in should have the same length as lat_temp/lon_temp/dates_use. 
; The model will update the local database file (if necessary) from the ECHAIM website daily, if the keyword 'update' is specified
; To force the DB file to be updated, use the keyword 'forceUpdate'

; NOTES: 
;       AE indices for 1988 and 1989 are spoofed using the PC index relationship. AE indices for 1976-1977 are spoofed using a PC index of 0.0.
;       The lat_temp, lon_temp, and dates_use variables can be vectors or single values but must all be of the same length.


; ----- EXAMPLE 1: Get profiles and parameters for a 24 hour period at 30 minute resolution:
;
;
; ALTS = 80.0+findgen(105)*5.0; Altitutdes array in km. 
; DATE = julday(07,21,2011,indgen(48)*0.5,00,00) ; every 30 minutes on July 21st, 2011
; LON = fltarr(48)+ 260.0 ;(can be either -180:180 or 0:360, all East)
; LAT = fltarr(48)+ 60.0 ; North Latitude
; output = ECHAIM(LAT,LON,DATE,ALTS,/Storm_nmF2); Storm NmF2 model is on.
; NmF2 = output.NmF2
; hmF2 = output.hmF2
; Dens_prof = output.dens ; The array of densities (N_times x N_alts).
; Alts_arr = output.alts ; An array of altitudes, provided for ease of contour plotting (N_times x N_alts). 
; juls_arr = output.juls ; An array of julian times, provided for ease of contour plotting(N_times x N_alts).
;
;; Plot NmF2:
; dummy = LABEL_DATE(DATE_FORMAT=['%H:%I'])
; plot, DATE,NmF2,psym = -2, XTICKUNITS = ['Hour'],XTICKFORMAT='LABEL_DATE'


; ----- EXAMPLE 2: Get in situ densities at a specified set of altitudes, locations, and times:
;
;
; ALTS = 400.0+findgen(105)*0.5; Vector of altitudes, 400:0.5:452.0
; DATE = julday(07,21,2011,00,indgen(105),00); One measurement every minute July 21st, 2011
; LON = findgen(105)*0.2+ 50.0; Track east at 0.2 degrees per minute
; LAT = findgen(105)*0.25+ 60.0; Track north at 0.1 degrees per minute
; output = ECHAIM(LAT,LON,DATE,ALTS,/Storm_nmF2, /Satellite) ; Storm NmF2 model is on.
; Dens = output.dens ; Vector of densities
;
;; Plot NmF2:
; dummy = LABEL_DATE(DATE_FORMAT=['%H:%I'])
; plot, DATE,Dens,psym = -2, XTICKUNITS = ['Hour'],XTICKFORMAT='LABEL_DATE'


; OUTPUT: 
;
; outs_struct - A structure returned through any call to ECHAIM_wrapper_NmF2_hmF2_combined_2018_new (the output variable in the above two examples) containing the following keywords:
;     MLON: AACGM Longitude corresponding to the input geographic locations.
;     MLAT: AACGM Latitude corresponding to the input geographic locations.
;     LTs: Local times corresponding to the input.
;     MLT: AACGM Magnetic Local Time corresponding to the input.
;     NmF2: Vector of NmF2 values in e/m^3.
;     hmF2: Vector of hmF2 values in km.
;     F107_27: 27-day smoothed F10.7 flux.
;     HBot: F2 region bottomside thickness in km (only for use with this model, no analogy to physical parameters).
;     HTop: Topside thickness at the F2 peak in km.
;     hmF1: Height of the F1 layer in km
;     HF1: Peak scale thickness associated with the F1 layer (only for use with this model, no analogy to physical parameters).
;     HE: Peak scale thickness associated with the E layer (only for use with this model, no analogy to physical parameters).
;     DENS: An array of output electron densities of dimension (N_times x N_altitudes).
;     ALTs: An array of output altitudes of dimension (N_times x N_altitudes). Provided for ease of contour plotting.
;     JULS: An array of output julian times of dimension (N_times x N_altitudes). Provided for ease of contour plotting.
;     AURORAL_Flux: Mean Auroral Flux used by the precipitation module (units of ergs/s/cm2)
;     AURORAL_E_0: Mean Auroral Energy used by the precipitation module (units of keV)
;     FLAGS - This is a string array that contains a series of error/warning codes that represent the following:
;                        A: Measured AE index was not available. A synthetic AE index based on PC index has been used instead for these data points.
;                        B: PC index is not available. Synthetic AE index is generated using PC = 0.0 for these data points.
;                        C: Data requested is beyond IG index record. IG12 has been used for this datapoint.
;                        E: These requested data periods are for a time past the maximum time stamp of the NOAA F10.7 flux forecast or prior to the first real F10.7 flux record. 
;                           F10.7 flux has been set to NAN. Output has been set to NAN
;                        F: The corresponding output was generated using a NOAA forecast of F10.7 flux: ftp://ftp.swpc.noaa.gov/pub/weekly/Predict.txt 
;                           Check the "NOAA F10.7 Prediction.txt" data file for the forecast date.
;                        G: Data requested is beyond or prior to IG12 index record. Output has been forced to NAN.
;                        H: Warning - Requested location is below the lower boundary of the model (50N geomagnetic). Since MLat is above 45N geomagnetic output is still provided.
;                           Output should be interpreted with caution.
;                        I: Warning - Requested location is below 45N geomagnetic latitude. Output has been forced to NAN.  
;                        J: Requested period is beyond or prior to the times available in the ap reference database. ap has been set to the all-time median (9.1)
;                        K: Requested period is beyond or prior to the times available in the dst reference database. dst has been set to the all-time median (-10.2) 
;
;

;Get the root directory
root_dir = file_dirname(routine_filepath()) + path_sep()

;get the version number
v = getVersion(root_dir)

; Print automatic disclaimer:

if ~keyword_set(silence) then begin
  if (!D.NAME eq 'WIN') then newline = string([13B, 10B]) else newline = string(10B)
  
  print, "Version v" + v + newline + "When using E-CHAIM for scientific purposes, please list the version number in any resulting publications.", $
          + newline + "Please consult the E-CHAIM Primer for full instructions." + newline + $
          "You may use the silence keyword to stop seeing this message." + newline
end


common AACGM_v2_Com, coef_v2,coefs_v2, cint_v2, $
  height_old_v2, order_v2,kmax_v2, $
  myear_v2, fyear_v2, myear_old_v2, fyear_old_v2, $
  aacgm_v2_datetime

;Update the DB file if requested
if keyword_set(update) or keyword_Set(forceUpdate) then begin
  
  r = updateLocalDB(keyword_set(forceUpdate))
  
  switch r of
    0: begin
        print, "The DB file has already been updated today."
        break
       end
    1: begin 
        print, "The website has not updated the DB file yet today."
        break
       end
    2: begin 
        print, "The DB file has been updated."
        break
       end
    3: begin 
        print, "The DB file was forced to update by the user."
        break
       end
  endswitch
  
endif

; ---Predefine output variables---:

NmF2_out_arr = fltarr(n_elements(lat_temp))
nmf2_upper_arr = fltarr(n_elements(lat_temp))
nmF2_lower_arr = fltarr(n_elements(lat_temp))
hours_lt_out_arr = fltarr(n_elements(lat_temp))
jul_lt_out = dblarr(n_elements(lat_temp))
hmF2_out_arr = fltarr(n_elements(lat_temp))
hmf2_upper_arr = fltarr(n_elements(lat_temp))
hmf2_lower_arr = fltarr(n_elements(lat_temp))
hmF1_out_arr = fltarr(n_elements(lat_temp))
Hbot_out_arr = fltarr(n_elements(lat_temp))
HE_out_arr = fltarr(n_elements(lat_temp))
HF1_out_arr = fltarr(n_elements(lat_temp))
Htop_out_arr = fltarr(n_elements(lat_temp))
Error_codes = strarr(n_elements(lat_temp))

if keyword_set(satellite) then begin
dens_out_arr = fltarr(n_elements(lat_temp))
dens_out_arr(0:*) = !values.F_NAN
alts_arr = fltarr(n_elements(lat_temp))
juls_arr = dblarr(n_elements(lat_temp))
endif else begin
dens_out_arr = fltarr(n_elements(lat_temp),n_elements(alts_in))
dens_out_arr(0:*) = !values.F_NAN
alts_arr = fltarr(n_elements(lat_temp),n_elements(alts_in))
juls_arr = dblarr(n_elements(lat_temp),n_elements(alts_in)) 
endelse

if keyword_set(precip) then begin
  out_energs = fltarr(n_elements(lat_temp))*!values.F_NAN
  out_efluxs = fltarr(n_elements(lat_temp))*!values.F_NAN
endif

pert_NmF2_out_arr = fltarr(n_elements(lat_temp))
pert_nmf2_upper_arr = fltarr(n_elements(lat_temp))
pert_nmF2_lower_arr = fltarr(n_elements(lat_temp))


; ---Define model parameter file locations---:

if !version.os_family eq 'Windows' then OS_flag = '\' else OS_flag = '/'

pert_dir = root_dir;
pert_file = 'perturb_model_save_file(May 7th 2018)(4-3-4 Base)(5-3 pert)(Update - Sept 21, 2018,5-3,just coefs).sav'

model_file = 'model_save_file(NmF2)(May-07-2018)(ALOG10 4-3-4)(Coefs Only).sav'
model_dir = root_dir;

hmF2_dir = root_dir;
hmF2_file = 'model_save_file(3-2-4)(hmF2)(2018)(Only Coefs).sav'

hmF1_file = 'model_save_file_hmF1(5-5-5)(With AE and SZA3)(2018)(Only Coefs).sav'
hmF1_dir = root_dir;

bottomside_dir = root_dir;
HBot_file = 'model_coefficients(H0)v2.sav';
HE_file = 'model_coefficients(HE)v2.sav';
HF1_file = 'model_coefficients(HF1)v2.sav';

topside_dir = root_dir;
topside_file = 'NeQuick_Topside_Model(5,5,4)(r = 20, g = 0.2024).sav'

AE_synth_dir = root_dir;
AE_synth_file = 'synth_AE_coefs.sav'

PC_index_dir = root_dir;
PC_index_file = 'PCIndex 2018-07-25.sav'

; ---Restore model coefficients from files and assign to new variables---:

restore, topside_dir+topside_file
H0Top_model_flapjacks = Chapjacks_1
H0Top_constants_out = chaps1_mod_constant_out

restore, bottomside_dir+HBot_file
HBot_model_flapjacks = model_flapjacks
HBot_constants_out = constants_out

restore, bottomside_dir+HE_file
HE_model_flapjacks = model_flapjacks
HE_constants_out = constants_out

restore, bottomside_dir+HF1_file
HF1_model_flapjacks = model_flapjacks
HF1_constants_out = constants_out

restore, hmF1_dir+hmF1_file
hmF1_model_flapjacks = model_flapjacks
hmF1_constants_out = constants_out

restore, hmF2_dir+hmF2_file
model_flapjacks_hmF2 = model_flapjacks
constants_out_hmF2 = constants_out 

if keyword_set(STORM_NmF2) then begin
restore, pert_dir+pert_file
model_flapjacks_pert = model_flapjacks
constants_out_pert = constants_out 
endif

restore, model_dir+model_file
constants_out_nmF2 = constants_out
model_flapjacks_nmF2 = model_flapjacks


; ---Calculations done at input time stamps---:

alts_use = fltarr(n_elements(lat_temp))
alts_use(0:*) = 300.0

caldat,dates_use,rescmonth,rescday,rescyear,reschour,rescminute,rescsecond
hour_lt = (julday(rescmonth,rescday,rescyear,reschour,rescminute,rescsecond)-julday(rescmonth,rescday,rescyear,00,00,00))*24.0

lt_struct_l = lt_and_mlat_calc_david_for_monthly_medians(lat_temp, lon_temp, alts_use,dates_use,root_dir)
resc_kp_l = lt_struct_l.J
resc_SSN_l = lt_struct_l.H
resc_F107_l = lt_struct_l.G
resc_sza_l_instant = lt_struct_l.F
resc_inclen_l = lt_struct_l.E
resc_modip_l = lt_struct_l.D
resc_mlon_l = lt_struct_l.C
resc_mlat_l = lt_struct_l.B
resc_jul_lt_l = lt_struct_l.K 
resc_hour_lt_l = lt_struct_l.A
resc_ig_l = lt_struct_l.L
resc_ig12_l = lt_struct_l.M
F107_l = lt_struct_l.Q
F107_81_l = lt_struct_l.N
F107_365_l = lt_struct_l.O
ap_8_l = lt_struct_l.ap8
MMLT_l =  lt_struct_l.P
geo_lat_use = lat_temp
geo_lon_use = lon_temp
lt_hour_use = (resc_jul_lt_l-floor(resc_jul_lt_l))*24.0
kp_str = lt_struct_l.kp_struct
dbDate = lt_struct_l.dbDate

; Track Errors:

errrs = lt_struct_l.errors
error_codes = error_codes + errrs

; Calculate Dipole tilts at input times:

dipole_tilts_instant = dipole_tilt_interface(dates_use)


; ---Load Geomagnetic Index Parameters---:

restore, root_dir + "IDL_DB.sav"

AE_int_ref = data.AEI
AE = data.AE
jul_int_AE_ref = data.hourly_jd

int_AE = interpol(AE_int_ref,jul_int_AE_ref,dates_use)
AE_l = interpol(AE,jul_int_AE_ref,dates_use)

int_PC = interpol(data.PCI,jul_int_AE_ref,dates_use)
PC_l = interpol(data.PC,jul_int_AE_ref,dates_use)


; ---Synthetic AE Index---:
a = where(finite(AE_int_ref))
max_ae = jul_int_AE_ref[a[-1]]
min_ae = jul_int_AE_ref[a[0]]

check_AE = where(dates_use gt max_ae or (rescyear eq 1988) or (rescyear eq 1989) or (rescyear eq 1976) or (rescyear eq 1977) or dates_use lt min_ae,cntaecheck)
if cntaecheck gt 0 then begin

Error_codes(check_AE) = Error_codes(check_AE)+'A'; Error code addition

;Set all PCI indices before Jan 1 1975 (and 1 hour) to 0 (start of PC index)
data.PCI[where(data.hourly_jd lt julday(1,1,1975,1,0,0))] = 0d0
; Integrate and interpolate PC index record:
int_PC_ref = [[data.PCI], [data.hourly_jd]]
k = where(finite(data.PCI))
int_PC = interpol(int_PC_ref(k,0),int_PC_ref(k,1),dates_use(check_AE),/nan)

; Find dates/years associated with periods of missing AE index:
temp_dates = dates_use(check_AE)
tempyears = rescyear(check_AE)

; Set PC index to 0.0 when PC is unavailable:
a = where(finite(data.PCI))
maxfinpc = data.hourly_jd[a[-1]]
minfinpc = data.hourly_jd[a[0]]
check_PC = where(temp_dates gt max(maxfinpc,/nan) or temp_dates lt (min(minfinpc,/nan)+2.0),cntpccheck) ; The +2.0 is to allow space for the integrater.
if cntpccheck gt 0 then begin
int_PC(check_PC) = 0.0
Error_codes(check_AE(check_PC)) = Error_codes(check_AE(check_PC))+'B' ; Error code addition
endif

; Restore Synthetic AE model parameters:
restore, AE_synth_dir+AE_synth_file
AE_constant_out = constant_out
AE_model_flapjacks = model_flapjacks

AE_aaaaaa = calc_synth_AE_model(dates_use(check_AE),F107_81_l(check_AE),dipole_tilts_instant(check_AE),abs(int_PC))

; Create synthetic AE index:
for flapp = 0L,cntaecheck-1L do begin
AE_model_par = transpose(AE_model_flapjacks)
int_AE(check_AE(flapp)) = (total(AE_aaaaaa((flapp),0:*)*AE_model_par)+AE_constant_out)
if int_AE(check_AE(flapp)) lt 15.0 then int_AE(check_AE(flapp)) = 15.0
endfor

endif


; ---Storm NmF2 model---:

if keyword_set(STORM_NmF2) then begin

ap_ref = kp_str.ap_hres
kp_ref_jul = kp_str.jul_hres

ap_param = [[data.API],[data.threehour_jd]]
ap_int_ref = ap_param(0:*,0)
jul_int_ap_ref = ap_param(0:*,1)
int_ap = interpol(ap_int_ref,jul_int_ap_ref,dates_use)

check_ap = where(dates_use gt max(jul_int_ap_ref(where(finite(ap_int_ref))),/nan) or dates_use lt min(jul_int_ap_ref(where(finite(ap_int_ref))),/nan),apcnt)
if apcnt gt 0 then begin
int_ap(check_ap) = 9.106;median(ap_int_ref); median value over entire reference dataset 
Error_codes(check_ap) = Error_codes(check_ap)+'J'
endif

int_dst_ref = [[data.DSTI],[data.hourly_jd]]
int_dst = interpol(int_dst_ref(0:*,0),int_dst_ref(0:*,1),dates_use)

refdstint = int_dst_ref(0:*,0)
refdstjul = int_dst_ref(0:*,1)

check_dst = where(dates_use gt max(refdstjul(where(finite(refdstint))),/nan) or dates_use lt min(refdstjul(where(finite(refdstint))),/nan),dstcnt)
if dstcnt gt 0 then begin
int_dst(check_dst) = -10.215;median(refdstint); median value over entire reference dataset 
Error_codes(check_dst) = Error_codes(check_dst)+'K'
endif

endif

; ---Create and index upper and lower times---:

bottom_hour = floor(hour_lt-0.5)+0.5
bottom_time = julday(rescmonth,rescday,rescyear,bottom_hour,00,00)
top_hour = ceil(hour_lt-0.5)+0.5
top_time = julday(rescmonth,rescday,rescyear,top_hour,00,00)


;---- Lower time calculations for hmF2 and NmF2----:

lt_struct = second_time_pars(lat_temp, lon_temp, alts_use,bottom_time)
resc_sza_l = lt_struct.F
dipole_tilts = dipole_tilt_interface(bottom_time)

aaarrr_ut_l = calc_CHAIM_model_parameters_for_NmF2_2018(resc_mlon_l,resc_mlat_l, bottom_time,resc_sza_l,resc_F107_l,resc_SSN_l,resc_modip_l,resc_inclen_l,resc_kp_l,resc_ig_l,resc_ig12_l,F107_81_l,F107_365_l,dipole_tilts,MMLT_l)
aaarrr_ut_l = aaarrr_ut_l.vars

aaarrr_ut_l_hmF2 = calc_CHAIM_model_parameters_for_lt_seperation_365_81_hmf2_2018(resc_mlon_l,resc_mlat_l, bottom_time,resc_sza_l,resc_F107_l,resc_SSN_l,resc_modip_l,resc_inclen_l,resc_kp_l,resc_ig_l,resc_ig12_l,F107_81_l,F107_365_l,dipole_tilts,MMLT_l)
aaarrr_ut_l_hmF2 = aaarrr_ut_l_hmF2.vars

if keyword_set(STORM_NmF2) then begin
aaarrr_ut_l_pert = calc_CHAIM_model_parameters_for_perturb_NmF2_2018(resc_mlon_l,resc_mlat_l, bottom_time,resc_sza_l,resc_F107_l,resc_SSN_l,resc_modip_l,resc_inclen_l,resc_ig_l,resc_ig12_l,F107_81_l,F107_365_l,int_dst,int_ap,dipole_tilts,int_AE,MMLT_l)
aaarrr_ut_l_pert = aaarrr_ut_l_pert.vars
endif

; --- hmF1, Hbot,HE, and HF1 calculations (i.e. parameters done only using input time stamps)---:


aaaaaa1_hmF1 = calc_CHAIM_model_parameters_for_hmF1(resc_mlon_l,resc_mlat_l, dates_use,resc_sza_l_instant,resc_F107_l,resc_SSN_l,resc_modip_l,resc_inclen_l,resc_kp_l,resc_ig_l,resc_ig12_l,F107_81_l,F107_365_l,dipole_tilts_instant,MMLT_l,lt_hour_use,geo_lat_use,jul_ut_use,int_AE)
aaarrr_ut_l_hmF1 = aaaaaa1_hmF1.vars

aaaaaa1_bottomside = calc_CHAIM_model_parameters_for_bottomside(resc_mlon_l,resc_mlat_l, dates_use,resc_sza_l_instant,resc_F107_l,resc_SSN_l,resc_modip_l,resc_inclen_l,resc_kp_l,resc_ig_l,resc_ig12_l,F107_81_l,F107_365_l,dipole_tilts_instant,MMLT_l,lt_hour_use,geo_lat_use,jul_ut_use,int_AE)
aaarrr_ut_l_bottomside = aaaaaa1_bottomside.vars

aaaaaa1_topside = calc_CHAIM_model_parameters_for_topside_2(resc_mlon_l,resc_mlat_l, dates_use,resc_sza_l_instant,resc_F107_l,resc_SSN_l,resc_modip_l,resc_inclen_l,resc_kp_l,resc_ig_l,resc_ig12_l,F107_81_l,F107_365_l,dipole_tilts_instant,MMLT_l,lt_hour_use,geo_lat_use,jul_ut_use,int_AE)
aaarrr_ut_l_topside = aaaaaa1_topside.vars


;---- Upper time for hmF2 and NmF2----:

lt_struct = second_time_pars(lat_temp, lon_temp, alts_use,top_time)
resc_sza = lt_struct.F
dipole_tilts = dipole_tilt_interface(top_time)

aaarrr_ut_u = calc_CHAIM_model_parameters_for_NmF2_2018(resc_mlon_l,resc_mlat_l, top_time,resc_sza,resc_F107_l,resc_SSN_l,resc_modip_l,resc_inclen_l,resc_kp_l,resc_ig_l,resc_ig12_l,F107_81_l,F107_365_l,dipole_tilts,MMLT_l)
aaarrr_ut_u = aaarrr_ut_u.vars

aaarrr_ut_u_hmF2 = calc_CHAIM_model_parameters_for_lt_seperation_365_81_hmf2_2018(resc_mlon_l,resc_mlat_l, top_time,resc_sza,resc_F107_l,resc_SSN_l,resc_modip_l,resc_inclen_l,resc_kp_l,resc_ig_l,resc_ig12_l,F107_81_l,F107_365_l,dipole_tilts,MMLT_l)
aaarrr_ut_u_hmF2 = aaarrr_ut_u_hmF2.vars

if keyword_set(STORM_NmF2) then begin
aaarrr_ut_u_pert = calc_CHAIM_model_parameters_for_perturb_NmF2_2018(resc_mlon_l,resc_mlat_l, top_time,resc_sza,resc_F107_l,resc_SSN_l,resc_modip_l,resc_inclen_l,resc_ig_l,resc_ig12_l,F107_81_l,F107_365_l,int_dst,int_ap,dipole_tilts,int_AE,MMLT_l)
aaarrr_ut_u_pert = aaarrr_ut_u_pert.vars
endif


; ---Calculation Loop---:

for fab = 0L, n_elements(lat_temp)-1L do begin

bottom_index = floor(hour_lt(fab)-0.5)
top_index = ceil(hour_lt(fab)-0.5)
if bottom_index lt 0 then bottom_index = 23
if top_index ge 24 then top_index = 0

bottom_hour = floor(hour_lt(fab)-0.5)+0.5
top_hour = ceil(hour_lt(fab)-0.5)+0.5
weight_top = abs(1.0-(top_hour-hour_lt(fab)))
weight_bottom = abs(1.0-(hour_lt(fab)-bottom_hour))


;-----NMF2----------:

flapjacks_lower = model_flapjacks_nmF2(bottom_index,0:*)
model_par_lower = transpose(flapjacks_lower)
NmF2_model_made_ut_lower = 10.0^(total(aaarrr_ut_l(fab,0:*)*model_par_lower)+constants_out_nmF2(bottom_index));^2.0

nmf2_lower_arr(fab) = NmF2_model_made_ut_lower

flapjacks_upper = model_flapjacks_nmF2(top_index,0:*)
model_par_upper = transpose(flapjacks_upper)
NmF2_model_made_ut_upper = 10.0^(total(aaarrr_ut_u(fab,0:*)*model_par_upper)+constants_out_nmF2(top_index));^2.0

nmf2_upper_arr(fab) = NmF2_model_made_ut_upper

NmF2_model_made_ut=((NmF2_model_made_ut_lower*weight_bottom) + (NmF2_model_made_ut_upper*weight_top))/(weight_bottom+weight_top)
jul_lt_out(fab) = (resc_jul_lt_l(fab))

NmF2_out_arr(fab) = NmF2_model_made_ut


;-----STORM NMF2----------:

if keyword_set(STORM_NmF2) then begin

flapjacks_lower = model_flapjacks_pert(bottom_index,0:*)
model_par_lower = transpose(flapjacks_lower)
NmF2_model_made_ut_lower = (total(aaarrr_ut_l_pert(fab,0:*)*model_par_lower)+constants_out_pert(bottom_index))

pert_nmf2_lower_arr(fab) = NmF2_model_made_ut_lower

flapjacks_upper = model_flapjacks_pert(top_index,0:*)
model_par_upper = transpose(flapjacks_upper)
NmF2_model_made_ut_upper = (total(aaarrr_ut_u_pert(fab,0:*)*model_par_upper)+constants_out_pert(top_index))

pert_nmf2_upper_arr(fab) = NmF2_model_made_ut_upper

pNmF2_model_made_ut=((NmF2_model_made_ut_lower*weight_bottom) + (NmF2_model_made_ut_upper*weight_top))/(weight_bottom+weight_top)
jul_lt_out(fab) = (resc_jul_lt_l(fab))

pert_NmF2_out_arr(fab) = pNmF2_model_made_ut

NmF2_out_arr(fab) = (10.0^pert_NmF2_out_arr(fab))*NmF2_out_arr(fab)
endif

;-----hmF2----------:

flapjacks_lower = model_flapjacks_hmF2(bottom_index,0:*)
model_par_lower = transpose(flapjacks_lower)
NmF2_model_made_ut_lower = (total(aaarrr_ut_l_hmF2(fab,0:*)*model_par_lower)+constants_out_hmF2(bottom_index))

hmf2_lower_arr(fab) = NmF2_model_made_ut_lower

flapjacks_upper = model_flapjacks_hmF2(top_index,0:*)
model_par_upper = transpose(flapjacks_upper)
NmF2_model_made_ut_upper = (total(aaarrr_ut_u_hmF2(fab,0:*)*model_par_upper)+constants_out_hmF2(top_index))

hmf2_upper_arr(fab) = NmF2_model_made_ut_upper

hmF2_model_made_ut=((NmF2_model_made_ut_lower*weight_bottom) + (NmF2_model_made_ut_upper*weight_top))/(weight_bottom+weight_top)
jul_lt_out(fab) = (resc_jul_lt_l(fab))

hmF2_out_arr(fab) = hmF2_model_made_ut


;----------hmF1, Bottomside, and Topside----------:

; hmF1:
hmF1_model_par_lower = transpose(hmF1_model_flapjacks)
if resc_sza_l_instant(fab) gt 102.0 then hmF1_out_arr(fab) = !values.F_NAN else hmF1_out_arr(fab) = (total(aaarrr_ut_l_hmF1(fab,0:*)*hmF1_model_par_lower)+hmF1_constants_out)

; HBot:
HBot_model_par_lower = transpose(HBot_model_flapjacks)
HBot_out_arr(fab) = (total(aaarrr_ut_l_bottomside(fab,0:*)*HBot_model_par_lower)+HBot_constants_out)

; HF1:
HF1_model_par_lower = transpose(HF1_model_flapjacks)
if resc_sza_l_instant(fab) gt 102.0 then HF1_out_arr(fab) = 0.0 else HF1_out_arr(fab) = (total(aaarrr_ut_l_bottomside(fab,0:*)*HF1_model_par_lower)+HF1_constants_out)
if HF1_out_arr(fab) lt 0.0 then HF1_out_arr(fab) = 0.0

; HE:
HE_model_par_lower = transpose(HE_model_flapjacks)
HE_out_arr(fab) = (total(aaarrr_ut_l_bottomside(fab,0:*)*HE_model_par_lower)+HE_constants_out)
if HE_out_arr(fab) lt 0.0 then HE_out_arr(fab) = 0.0

; Htop:
Htop_model_par_lower = transpose(H0Top_model_flapjacks)
Htop_out_arr(fab) = (total(aaarrr_ut_l_topside(fab,0:*)*Htop_model_par_lower)+H0Top_constants_out)

; ---Check for NmF2 or hmF2 user input---:

if keyword_set(user_hmF2) then hmF2_out_arr = user_hmF2
if keyword_set(user_NmF2) then NmF2_out_arr = user_NmF2


;-----------Generate Satellite------:

if keyword_set(satellite) then begin

sat_alt = alts_in(fab)

check_alts = where(sat_alt le hmF2_out_arr(fab), cnt_alts)
if cnt_alts gt 0 then begin
bot_dens = NmF2_out_arr(fab)*ECHAIM_bottomside_function(sat_alt(check_alts),hmF2_out_arr(fab),hmF1_out_arr(fab),102.0,HBot_out_arr(fab),HF1_out_arr(fab),HE_out_arr(fab))
dens_out_arr(fab,check_alts) = bot_dens
endif

check_alts = where(sat_alt gt hmF2_out_arr(fab), cnt_alts)
if cnt_alts gt 0 then begin
top_dens = NmF2_out_arr(fab)*nequick_model_new_r_g(sat_alt(check_alts)-hmF2_out_arr(fab),Htop_out_arr(fab))
dens_out_arr(fab,check_alts) = top_dens
endif

alts_arr(fab,0:*) = sat_alt 
juls_arr(fab,0:*) = dates_use(fab)

endif else begin

;-----------Generate Profiles------:

check_alts = where(alts_in le hmF2_out_arr(fab), cnt_alts)
if cnt_alts gt 0 then begin
bot_dens = NmF2_out_arr(fab)*ECHAIM_bottomside_function(alts_in(check_alts),hmF2_out_arr(fab),hmF1_out_arr(fab),102.0,HBot_out_arr(fab),HF1_out_arr(fab),HE_out_arr(fab))
dens_out_arr(fab,check_alts) = bot_dens
endif

check_alts = where(alts_in gt hmF2_out_arr(fab), cnt_alts)
if cnt_alts gt 0 then begin
top_dens = NmF2_out_arr(fab)*nequick_model_new_r_g(alts_in(check_alts)-hmF2_out_arr(fab),Htop_out_arr(fab))
dens_out_arr(fab,check_alts) = top_dens
endif

alts_arr(fab,0:*) = alts_in 
juls_arr(fab,0:*) = dates_use(fab)

endelse ; keyword satellite else statement

;D-region module
if keyword_set(dregion) then begin
  
  ;get DOY
  caldat,dates_use[fab],month,day,year
  doy = 1 + julday(month,day,year) - julday(1,1,year)
  
  ;for single altitudes
  if keyword_set(satellite) then begin
    pert = ECHAIM_FIRI2018(root_dir, lat_temp(fab), resc_sza_l_instant[fab], F107_l[fab], doy, alts_in[fab])
    dens_out_arr[fab] += pert
  endif else begin
    ;get the perturbations for each altitude
    for aa = 0, n_elements(alts_in)-1 do begin
      pert = ECHAIM_FIRI2018(root_dir, lat_temp(fab), resc_sza_l_instant[fab], F107_l[fab], doy, alts_in[aa])
      dens_out_arr[fab,aa] += pert
    endfor
  endelse

end

; ---Precipitation Module---:

if keyword_set(precip) then begin

  ;set required input parameters
  F107 = F107_l[fab]
  F107_81 = F107_81_l[fab]
  Ap_l = ap_8_l[fab]
  
  if finite(AE_l[fab]) then begin
    AE = AE_l[fab]
    isae = 1
  endif else begin
    AE = PC_l[fab]
    isae = 0
  endelse
  
  ;check if PC is also missing
  if finite(AE) eq 0 then begin
    AE = 117.7 ;set to the mean AE
    isae = 1
  endif

  erg2ev = 6.2415e11 ;convert erg to ev
  mlat = resc_mlat_l[fab]
  mlt = MMLT_l[fab]
  glat = lat_temp[fab]
  glon = lon_temp[fab]
  caldat,dates_use[fab],months,days,years,hours,minutes,seconds
  utc = hours+minutes/60.0+seconds/3600.0
  date = string(months,format = '(I02)')+ string(days,format = '(I02)')+string(years,format = '(I04)')
  
  ;set proper alts
  if keyword_set(satellite) then begin
    alts = alts_in(fab)
  endif else begin
    alts = alts_in
  endelse
  
  ;mean energy and flux
  energyFluxModel,mlt, mlat, AE, isae, root_dir, ener_mean, eflux_mean
  
  ;check for negative values
  if ener_mean lt 0d0 then ener_mean = 0d0
  if eflux_mean lt 0d0 then eflux_mean = 0d0
  
  ; Store energies and fluxes for user output:
  out_energs(fab) = ener_mean
  out_efluxs(fab) = eflux_mean
  
  if eflux_mean gt 0.05 then begin
  
    ener_mean /= 2.0
    eflux_mean *= erg2ev / 1000.0 ;convert flux from erg to ev
    ener_mean = ener_mean[0];
    eflux_mean = eflux_mean[0];
    
    ;check that the altitudes are less than 400, otherwise nothing is applied
    whereAlts = where(alts lt 400.0)
    if whereAlts[0] eq -1 then continue ;continue the calculation loop, leaving this density output unmodified
  
    ;get the atmospheric data from NRLMSISE
    atm_data=nrlmsise00_arr(date,utc,alts(whereAlts),glat,glon,f107_daily=F107,f107_3month=F107_81,ap_daily=Ap_l)
    
    ;generate energy spectrum
    ;espec=energy_spectrum(ener_mean,1,[0.1,400.0],n_elem=1000,log_op=1,kappa_op=0)
    espec=gauss_energy_spectrum(ener_mean, w)
    norm_flux=espec.flux
    ener_arr=espec.energy
    qh2=alts(whereAlts)
    q_arr=fltarr(n_elements(ener_arr),n_elements(qh2))
    qtot=fltarr(n_elements(qh2))
    for ie=0,n_elements(ener_arr)-1 do begin
      imono_rate=fast_eion_rate(glat,ener_arr[ie],norm_flux[ie]*eflux_mean,atm_data)
      q_arr[ie,*]=imono_rate
    endfor
    
    for ih=0,n_elements(qh2)-1 do begin
      iqarr=q_arr[*,ih]
      qtot[ih]=total(iqarr*w);integrate spectrum
    endfor
    alpha=4.3e-6*exp(-2.42e-2*qh2)+8.16e12*exp(-0.524*qh2); define Gledhill recombination rates
    aaff = alpha
    cuth = 130.0
    aaffs = where(qh2 gt cuth, cntaaffs)
    
    ; Set Gledhill recombinations to constant above threshold altitude to suppress F-region precipitation artifacts.
    if cntaaffs gt 0 then aaff(aaffs) = 4.3e-6*exp(-2.42e-2*cuth)+8.16e12*exp(-0.524*cuth)
    sigmoid = 1.0/(1.0+exp((cuth-qh2)/40.0))
    alpha2 = alpha*(1.0-sigmoid)+(sigmoid)*aaff
    
    ; Calculate Q/aeff:
    Qtot_alpha = Qtot/alpha2
    
    k = where(finite(Qtot_alpha) eq 0)
    if k[0] ne -1 then Qtot_alpha[k] = 0.0
    
    dens_out_arr(fab,whereAlts) = sqrt((dens_out_arr(fab,whereAlts)/1e6)^2.0+Qtot_alpha)*1e6; Determine new electron density profile.

  endif
endif

endfor ; End of calculation loop

; Error code and reject requests below lower boundary (Warning between 50N and 45N, NAN below 45N):
boundarys = where(resc_mlat_l ge 45.0 and resc_mlat_l lt 50.0, cntbound)
if cntbound gt 0 then begin
Error_codes(boundarys) = Error_codes(boundarys)+'H'
endif 

boundarys2 = where(resc_mlat_l lt 45.0, cntbound2)
if cntbound2 gt 0 then begin
NmF2_out_arr(boundarys2) = !values.F_NAN
hmF2_out_arr(boundarys2) = !values.F_NAN
hmF1_out_arr(boundarys2) = !values.F_NAN
HBot_out_arr(boundarys2) = !values.F_NAN
HF1_out_arr(boundarys2) = !values.F_NAN
HE_out_arr(boundarys2) = !values.F_NAN
dens_out_arr(boundarys2,0:*) = !values.F_NAN
Htop_out_arr(boundarys2) = !values.F_NAN
Error_codes(boundarys2) = Error_codes(boundarys2)+'I'
endif 

; define output structure:
outs_struct = {MLON:resc_mlon_l, MLAT:resc_mlat_l, NmF2:NmF2_out_arr, LTs:jul_lt_out,MLT:MMLT_l, hmF2:hmF2_out_arr, F107_27: resc_F107_l,hmF1:hmF1_out_arr, HBot:HBot_out_arr, HF1:HF1_out_arr, HE:HE_out_arr, DENS:dens_out_arr, ALTS:alts_arr, JULS:juls_arr, HTop:Htop_out_arr,FLAGS:Error_codes}
if keyword_set(precip) then outs_struct = {MLON:resc_mlon_l, MLAT:resc_mlat_l, NmF2:NmF2_out_arr, LTs:jul_lt_out,MLT:MMLT_l, hmF2:hmF2_out_arr, F107_27: resc_F107_l,hmF1:hmF1_out_arr, HBot:HBot_out_arr, HF1:HF1_out_arr, HE:HE_out_arr, DENS:dens_out_arr, ALTS:alts_arr, JULS:juls_arr, HTop:Htop_out_arr,FLAGS:Error_codes,Auroral_E_0:out_energs, Auroral_Flux: out_efluxs}

;set metadata
outs_struct = create_Struct(outs_Struct, 'dbDate', dbDate)

return, outs_struct

end